import typing

from retrack.nodes.base import NodeKind
import pydantic

from retrack.nodes.base import InputConnectionModel, OutputConnectionModel
from retrack.nodes.dynamic.base import BaseDynamicIOModel, BaseDynamicNode


class FlowConnectorMetadataModel(pydantic.BaseModel):
    name: str
    flow: str
    version: str
    default: typing.Optional[str] = None


class FlowConnectorOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


def flow_connector_factory(
    inputs: typing.Dict[str, typing.Any], **kwargs
) -> typing.Type[BaseDynamicNode]:
    input_fields = {}

    for name in inputs.keys():
        input_fields[name] = BaseDynamicNode.create_sub_field(InputConnectionModel)

    inputs_model = BaseDynamicIOModel.with_fields(
        "FlowConnectorInputsModel", **input_fields
    )

    models = {
        "inputs": BaseDynamicNode.create_sub_field(inputs_model),
        "outputs": BaseDynamicNode.create_sub_field(FlowConnectorOutputsModel),
        "data": BaseDynamicNode.create_sub_field(FlowConnectorMetadataModel),
    }

    BaseModel = BaseDynamicNode.with_fields("FlowConnectorBaseModel", **models)

    class FlowConnector(BaseModel):
        def kind(self) -> NodeKind:
            return NodeKind.CONNECTOR

        async def run(self, **kwargs):
            return {}

    return FlowConnector



from typing import Optional
import networkx as nx

from retrack.validators.base import BaseValidator


class CheckIsDAG(BaseValidator):
    def validate(self, edges: list, **kwargs) -> tuple[bool, Optional[str]]:
        """Validate the graph data.

        Returns:
            A tuple (is_valid, error_message). is_valid is True if the graph data is valid, False otherwise.
            error_message contains details about the validation failure, or None if validation passed.
        """
        graph = nx.DiGraph()
        graph.add_edges_from(edges)
        result = nx.is_directed_acyclic_graph(graph)
        return result, None if result else "Graph is not a DAG"



import io
import typing

import numpy as np
import pandas as pd
import pydantic

from retrack.nodes.base import (
    BaseNode,
    InputConnectionModel,
    NodeKind,
    NodeMemoryType,
    OutputConnectionModel,
)

#######################################################
# Constant Metadata Models
#######################################################


class ConstantMetadataModel(pydantic.BaseModel):
    value: str
    name: typing.Optional[str] = None


class ListMetadataModel(pydantic.BaseModel):
    value: typing.List[str]


class BoolMetadataModel(pydantic.BaseModel):
    value: typing.Optional[bool] = pydantic.Field(False, alias="value")

    @pydantic.field_validator("value")
    def validate_value(cls, value):
        if value is None:
            return False

        value = str(value).lower()

        if value in ["true", "1", "yes", "y", "t"]:
            return True

        return False


class IntervalCatMetadataModel(pydantic.BaseModel):
    value: typing.List[str]
    start_interval_column: str
    end_interval_column: str
    category_column: str
    headers: typing.List[str]
    separator: typing.Optional[str] = pydantic.Field(",")
    default: typing.Optional[str] = None

    def df(self) -> pd.DataFrame:
        csv_str = "\n".join(self.value[1:])
        buffer = io.StringIO(csv_str)

        df = pd.read_csv(
            buffer,
            header=None,
            names=self.headers,
            sep=self.separator,
            dtype={
                self.start_interval_column: "float64",
                self.end_interval_column: "float64",
            },
        )

        return df


#######################################################
# Constant Inputs and Outputs
#######################################################


class ConstantInputsModel(pydantic.BaseModel):
    input_void: typing.Optional[InputConnectionModel] = None


class ConstantInputsValueModel(pydantic.BaseModel):
    input_value: InputConnectionModel


class ConstantOutputsModel(pydantic.BaseModel):
    output_value: OutputConnectionModel


class ListOutputsModel(pydantic.BaseModel):
    output_list: OutputConnectionModel


class BoolOutputsModel(pydantic.BaseModel):
    output_bool: OutputConnectionModel


#######################################################
# Constant Nodes
#######################################################


class BaseConstant(BaseNode):
    inputs: typing.Optional[ConstantInputsModel] = None

    def kind(self) -> NodeKind:
        return NodeKind.CONSTANT


class Constant(BaseConstant):
    data: ConstantMetadataModel
    outputs: ConstantOutputsModel

    async def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {"output_value": self.data.value}


class List(BaseConstant):
    data: ListMetadataModel
    outputs: ListOutputsModel

    async def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {}  # {"output_list": self.data.value}

    def memory_type(self) -> NodeMemoryType:
        return NodeMemoryType.CONSTANT


class Bool(BaseConstant):
    data: BoolMetadataModel = BoolMetadataModel(value=False)
    outputs: BoolOutputsModel

    async def run(self, **kwargs) -> typing.Dict[str, typing.Any]:
        return {"output_bool": self.data.value}


class IntervalCatV0(BaseConstant):
    data: IntervalCatMetadataModel
    inputs: ConstantInputsValueModel
    outputs: ConstantOutputsModel

    async def run(self, input_value: pd.Series) -> typing.Dict[str, typing.Any]:
        values = pd.to_numeric(input_value, errors="coerce").to_numpy()

        df = self.data.df()
        starts = pd.to_numeric(
            df[self.data.start_interval_column], errors="coerce"
        ).to_numpy()
        ends = pd.to_numeric(
            df[self.data.end_interval_column], errors="coerce"
        ).to_numpy()
        cats_s = df[self.data.category_column]
        intervals = pd.IntervalIndex.from_arrays(starts, ends, closed="left")
        idx = intervals.get_indexer(values)
        cats = cats_s.to_numpy(dtype=object)
        out = np.full(values.shape, np.nan, dtype=object)

        found_mask = idx != -1
        if found_mask.any():
            mapped = cats[idx[found_mask]]
            out[found_mask] = mapped

            pass_mask = mapped == "{value}"
            if pass_mask.any():
                ii = np.where(found_mask)[0][pass_mask]
                out[ii] = input_value.iloc[ii]

        if self.data.default is not None:
            nan_mask = pd.isna(out)
            if nan_mask.any():
                out[nan_mask] = self.data.default

        output = pd.Series(out, index=input_value.index, dtype="object")

        return {"output_value": output}



import typing

import numpy as np
import pandas as pd

from retrack.nodes.base import BaseNode
from retrack.utils import constants, registry
from retrack.engine.schemas import ExecutionSchema
from retrack.utils.transformers import (
    serialize_connections,
    explode_nodes_by_values_iter,
    normalize_execution_for_debug_iter,
    to_metadata,
)


class Execution:
    def __init__(
        self,
        payload: pd.DataFrame,
        states: pd.DataFrame,
        filters: dict = None,
        context: registry.Registry = None,
        nodes: dict = None,
        constants: dict = None,
    ):
        self.payload = payload
        self.states = states
        self.filters = filters or {}
        self.context = context
        self.nodes = nodes or {}
        self.constants = constants or {}

    def set_state_data(
        self, column: str, value: typing.Any, filter_by: typing.Any = None
    ):
        if filter_by is None:
            self.states[column] = value
        else:
            self.states.loc[filter_by, column] = value

    def get_state_data(
        self, column: str, constants: dict, filter_by: typing.Any = None
    ):
        if column in constants:
            return constants[column]

        if filter_by is None:
            return self.states[column]

        return self.states.loc[filter_by, column]

    def set_constants_data(self, constants: dict):
        self.constants.update(constants)

    def add_node(self, node: BaseNode):
        self.nodes[node.id] = node

    def update_filters(self, filter_value, output_connections: typing.List[str] = None):
        for output_connection_id in output_connections:
            if self.filters.get(output_connection_id, None) is None:
                self.filters[output_connection_id] = filter_value
            else:
                self.filters[output_connection_id] = (
                    self.filters[output_connection_id] & filter_value
                )

    @classmethod
    def from_payload(
        cls,
        validated_payload: pd.DataFrame,
        input_columns: dict,
        context: registry.Registry = None,
    ):
        state_df = pd.DataFrame([])
        for node_id, input_name in input_columns.items():
            state_df[node_id] = validated_payload[input_name].copy()

        state_df[constants.OUTPUT_REFERENCE_COLUMN] = np.nan
        state_df[constants.OUTPUT_MESSAGE_REFERENCE_COLUMN] = np.nan

        return cls(
            payload=validated_payload,
            states=state_df,
            context=context,
        )

    @property
    def result(self) -> pd.DataFrame:
        return self.states[
            [
                constants.OUTPUT_REFERENCE_COLUMN,
                constants.OUTPUT_MESSAGE_REFERENCE_COLUMN,
            ]
        ]

    def has_ended(self) -> bool:
        return self.states[constants.OUTPUT_REFERENCE_COLUMN].isna().sum() == 0

    def to_dict(self) -> dict:
        return {
            "payload": self.payload.to_dict(),
            "states": self.states.to_dict(),
            "filters": {k: v.to_dict() for k, v in self.filters.items()},
            "result": self.result.to_dict(),
            "has_ended": self.has_ended(),
        }

    def to_model(self) -> ExecutionSchema:
        return ExecutionSchema(**self.to_dict())

    def to_normalized_dict(self) -> list:
        nodes_normalized = []
        values_cache = {}
        for node in self.nodes.values():
            nodes_normalized.append(
                {
                    "id": node.id,
                    "name": node.alias(),
                    "type": node.type(),
                    "inputs": serialize_connections(
                        node.inputs,
                        node_id=node.id,
                        connection_type="input",
                        execution=self,
                        values_cache=values_cache,
                    ),
                    "outputs": serialize_connections(
                        node.outputs,
                        node_id=node.id,
                        connection_type="output",
                        execution=self,
                        values_cache=values_cache,
                    ),
                    "default": node.default(),
                    "data": to_metadata(node),
                }
            )

        exploded_nodes = explode_nodes_by_values_iter(nodes_normalized)
        return list(normalize_execution_for_debug_iter(exploded_nodes))

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            payload=pd.DataFrame(data["payload"]),
            states=pd.DataFrame(data["states"]),
            filters={k: pd.DataFrame(v) for k, v in data["filters"].items()},
        )

    def __repr__(self) -> str:
        return f"Execution({self.to_dict()})"

    def __str__(self) -> str:
        return self.__repr__()
